from typing import Callable, Dict, Tuple, Type

from tdm.future.abstract.datamodel import AbstractNodeMention
from tdm.future.abstract.json_schema import ElementModel, get_model_generator


def register_mention_models() -> \
        Tuple[Dict[Type[AbstractNodeMention], Type[ElementModel]], Callable[[AbstractNodeMention], ElementModel]]:
    import tdm.future.datamodel.mentions as mentions  # we need mentions models to be registered
    mentions

    # TODO: add plugins support here

    return get_model_generator(AbstractNodeMention).generate_union_model(discriminator='type')


MENTION_MODELS, serialize_mention = register_mention_models()



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterable, Optional, Set, Tuple, Type, TypeVar, Union

from .directive import AbstractDirective
from .fact import AbstractFact
from .link import AbstractNodeLink
from .node import AbstractNode

_TD = TypeVar('_TD', bound='TalismanDocument')

NodeOrId = Union[str, AbstractNode]


class TalismanDocument(metaclass=ABCMeta):
    @property
    @abstractmethod
    def id(self) -> str:
        pass

    @property
    @abstractmethod
    def nodes(self) -> Dict[str, AbstractNode]:
        pass

    @abstractmethod
    def with_nodes(self: _TD, nodes: Iterable[AbstractNode]) -> _TD:
        pass

    @abstractmethod
    def without_nodes(self: _TD, nodes: Iterable[NodeOrId], *, cascade: bool = True) -> _TD:
        pass

    @property
    @abstractmethod
    def roots(self) -> Set[AbstractNode]:
        pass

    @property
    @abstractmethod
    def main_root(self) -> Optional[AbstractNode]:
        pass

    @abstractmethod
    def with_main_root(self: _TD, node: Optional[NodeOrId], *, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def parent(self, node: NodeOrId) -> Optional[AbstractNode]:
        pass

    @abstractmethod
    def child_nodes(self, node: Union[str, AbstractNode]) -> Tuple[AbstractNode, ...]:
        pass

    @abstractmethod
    def with_links(self: _TD, links: Dict[NodeOrId, Iterable[NodeOrId]], *, force: bool = False, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def with_link(self: _TD, parent: NodeOrId, child: NodeOrId, *, force: bool = False, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def without_parent_links(self: _TD, nodes: Iterable[NodeOrId]) -> _TD:
        pass

    @abstractmethod
    def without_links(self: _TD, links: Dict[NodeOrId, Iterable[NodeOrId]]) -> _TD:
        pass

    @property
    @abstractmethod
    def node_links(self) -> Dict[Type[AbstractNodeLink], Iterable[AbstractNodeLink]]:
        pass

    @abstractmethod
    def with_node_links(self: _TD, links: Iterable[AbstractNodeLink], *, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def without_node_links(self: _TD, links: Iterable[Union[str, AbstractNodeLink]], *, cascade: bool = False) -> _TD:
        pass

    @property
    @abstractmethod
    def facts(self) -> Dict[Type[AbstractFact], Iterable[AbstractFact]]:
        pass

    @abstractmethod
    def with_facts(self: _TD, facts: Iterable[AbstractFact], *, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def without_facts(self: _TD, facts: Iterable[Union[str, AbstractFact]], *, cascade: bool = False) -> _TD:
        pass

    @property
    @abstractmethod
    def directives(self) -> Dict[Type[AbstractDirective], AbstractDirective]:
        pass

    @abstractmethod
    def with_directives(self: _TD, directives: Iterable[AbstractDirective], *, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def without_directives(self: _TD, directives: Iterable[Union[str, AbstractDirective]], *, cascade: bool = False) -> _TD:
        pass

    @property
    @abstractmethod
    def metadata(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def with_metadata(self: _TD, metadata: Dict[str, Any]) -> _TD:
        pass


class AbstractDocumentFactory(metaclass=ABCMeta):
    @abstractmethod
    def create_document(self, *, id_: Optional[str] = None) -> TalismanDocument:
        pass



__all__ = [
    'EnsureIdentifiable', 'Identifiable',
    'AbstractDirective',
    'AbstractDocumentFactory', 'TalismanDocument',
    'AbstractFact', 'AbstractLinkFact', 'FactStatus',
    'AbstractNodeLink',
    'AbstractNodeMention',
    'AbstractContentNode', 'AbstractNode', 'BaseNodeMetadata'
]

from .base import EnsureIdentifiable, Identifiable
from .directive import AbstractDirective
from .document import AbstractDocumentFactory, TalismanDocument
from .fact import AbstractFact, AbstractLinkFact, FactStatus
from .link import AbstractNodeLink
from .mention import AbstractNodeMention
from .node import AbstractContentNode, AbstractNode, BaseNodeMetadata



from dataclasses import dataclass
from typing import Set

from .base import EnsureIdentifiable


@dataclass(frozen=True)
class AbstractDirective(EnsureIdentifiable):

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return set()



from typing import Dict, Optional, Tuple

from pydantic import BaseModel, root_validator
from typing_extensions import Literal

from tdm.future.abstract.datamodel import AbstractNode, TalismanDocument
from tdm.future.datamodel.document import TalismanDocumentFactory
from tdm.json_schema import DocumentMetadataModel
from .directives import DirectivesModel
from .facts import FactsModel
from .links import NodeLinksModel
from .nodes import NodeModel, fill_children, serialize_node


class TalismanDocumentModel(BaseModel):
    VERSION: Literal['1.0'] = ...
    id: str
    main_node: str
    content: Tuple[NodeModel, ...]
    links: NodeLinksModel
    facts: FactsModel
    directives: DirectivesModel

    metadata: Optional[DocumentMetadataModel]  # should be removed in future

    @root_validator(pre=True)
    def _set_version(cls, values):  # noqa N805: pydantic requires cls to be first argument in validators
        if 'VERSION' not in values:
            values['VERSION'] = '1.0'
        return values

    def deserialize(self) -> TalismanDocument:
        id2node, structure = self._collect_nodes()

        return TalismanDocumentFactory.construct(
            content=id2node.values(),
            structure=structure,
            root=self.main_node,
            node_links=self.links.deserialize(id2node),
            facts=self.facts.deserialize(id2node),
            directives=self.directives.deserialize(id2node),
            metadata=self.metadata.to_metadata() if self.metadata is not None else None,
            id_=self.id
        )

    def _collect_nodes(self) -> Tuple[Dict[str, AbstractNode], Dict[str, Tuple[str, ...]]]:
        # try to avoid recursion
        links: Dict[str, Tuple[str, ...]] = {}
        id2node = {}

        for node_model in self.content:
            id2node[node_model.id] = node_model.deserialize({})
            if node_model.children:
                links[node_model.id] = node_model.children
        return id2node, links

    @classmethod
    def serialize(cls, document: TalismanDocument) -> 'TalismanDocumentModel':
        node_models = tuple(serialize_node(node) for node in document.nodes.values())
        fill_children(node_models, document)
        return cls(
            id=document.id,
            main_node=document.main_root.id,
            content=node_models,
            links=NodeLinksModel.serialize(document.node_links),
            facts=FactsModel.serialize(document.facts),
            directives=DirectivesModel.serialize(document.directives),
            metadata=DocumentMetadataModel(**document._metadata) if document._metadata is not None else None
        )



from copy import deepcopy
from functools import wraps
from typing import Any, Dict, Iterable, Iterator, Optional, Set, Tuple, Type, TypeVar, Union

from tdm.abstract.datamodel.document import DocumentMetadata
from tdm.future.abstract.datamodel import AbstractDirective, AbstractFact, AbstractNode, AbstractNodeLink, EnsureIdentifiable, \
    TalismanDocument
from ._base import BaseImpl
from ._container import TypedIdsContainer
from ._structure import NodesStructure
from ._view import AbstractView, restore_object

_TD = TypeVar('_TD', bound='TalismanDocumentImpl')
_D = TypeVar('_D', bound=EnsureIdentifiable)


def _filter_elements(base_type: Type[_D]):
    def decorator(f):
        @wraps(f)
        def wrapper(self: _TD, els: Iterable[Union[str, _D]], *args, **kwargs):
            return f(self, _filter(self._containers[base_type], self._id2view, els, base_type), *args, **kwargs)

        return wrapper

    return decorator


class TalismanDocumentImpl(TalismanDocument, BaseImpl):
    __slots__ = (
        '_id', '_structure', '_metadata'
    )

    def __init__(
            self,
            id2view: Dict[str, Union[EnsureIdentifiable, AbstractView]],
            dependencies: Dict[str, Set[Tuple[str, Type[EnsureIdentifiable]]]],
            containers: Dict[Type[EnsureIdentifiable], TypedIdsContainer],
            structure: NodesStructure,
            metadata: Optional[DocumentMetadata],  # should be removed in future
            *, id_: str):
        super().__init__(id2view, dependencies, containers)
        self._id = id_
        self._structure = structure

        self._metadata = deepcopy(metadata)  # should be removed in future

    def _replace(self: _TD, **kwargs) -> _TD:
        return super()._replace(
            **{
                'structure': self._structure,
                'metadata': self._metadata,
                'id_': self._id,
                **kwargs
            }
        )

    def _update_structure(self: _TD, other: _TD) -> _TD:
        if other._containers[AbstractNode] is not self._containers[AbstractNode]:  # something has changed
            other._structure = self._structure.update_nodes(other._containers[AbstractNode].ids)
        return other

    def with_elements(self: _TD, elements: Iterable[EnsureIdentifiable], *, update: bool = False) -> _TD:
        return self._update_structure(super().with_elements(elements, update=update))

    def without_elements(self: _TD, ids: Iterable[str], *, cascade: bool = False) -> _TD:
        return self._update_structure(super().without_elements(ids, cascade=cascade))

    @property
    def id(self) -> str:
        return self._id

    @property
    def nodes(self) -> Dict[str, AbstractNode]:
        return {id_: restore_object(self._id2view[id_], self._id2view) for id_ in self._containers[AbstractNode].ids}

    @_filter_elements(AbstractNode)
    def with_nodes(self: _TD, nodes: Iterable[AbstractNode]) -> _TD:
        return self.with_elements(nodes, update=True)

    @_filter_elements(AbstractNode)
    def without_nodes(self: _TD, nodes: Iterable[Union[str, AbstractNode]], *, cascade: bool = False) -> _TD:
        return self.without_elements(map(_to_id, nodes), cascade=cascade)

    @property
    def roots(self) -> Set[AbstractNode]:
        return {restore_object(self._id2view[id_], self._id2view) for id_ in self._structure.roots}

    @property
    def main_root(self) -> Optional[AbstractNode]:
        main_root = self._structure.main_root
        return restore_object(self._id2view[main_root], self._id2view) if main_root is not None else None

    def with_main_root(self: _TD, node: Optional[Union[str, AbstractNode]], *, update: bool = False) -> _TD:
        if node is None:
            return self._replace(structure=self._structure.with_main_root(None))
        if update:
            if not isinstance(node, AbstractNode):
                raise ValueError(f"could not update node {node} with id only")
            result = self.with_nodes((node,))
            result._structure = result._structure.with_main_root(node.id)
            return result
        return self._replace(structure=self._structure.with_main_root(_to_id(node)))

    def parent(self, node: Union[str, AbstractNode]) -> Optional[AbstractNode]:
        node_id = _to_id(node)
        parent_id = self._structure.parent.get(node_id)
        if parent_id is not None:
            return restore_object(self._id2view[parent_id], self._id2view)
        return None

    def child_nodes(self, node: Union[str, AbstractNode]) -> Tuple[AbstractNode, ...]:
        return tuple(restore_object(self._id2view[id_], self._id2view) for id_ in self._structure.children.get(_to_id(node), ()))

    def with_links(
            self: _TD,
            links: Dict[Union[str, AbstractNode], Iterable[Union[str, AbstractNode]]],
            *, force: bool = False, update: bool = False
    ) -> _TD:
        structure = {}
        nodes = set()

        for parent, children in links.items():
            children = list(children)
            if update:
                if isinstance(parent, AbstractNode):
                    nodes.add(parent)
                nodes.update(node for node in children if isinstance(node, AbstractNode))
            structure[_to_id(parent)] = [_to_id(node) for node in children]

        result = self
        if update:
            result = self.with_nodes(nodes)

        return result._replace(structure=result._structure.with_children(structure, force=force))

    def with_link(
            self: _TD,
            parent: Union[str, AbstractNode], child: Union[str, AbstractNode],
            *, force: bool = False, update: bool = False
    ) -> _TD:
        result = self

        if update:
            nodes = {node for node in (child, parent) if isinstance(node, AbstractNode)}
            result = result.with_nodes(nodes)

        return result._replace(structure=result._structure.with_parent(_to_id(parent), _to_id(child), force=force))

    def without_parent_links(self: _TD, nodes: Iterable[Union[str, AbstractNode]]) -> _TD:
        return self._replace(structure=self._structure.as_roots(map(_to_id, nodes)))

    def without_links(self: _TD, links: Dict[Union[str, AbstractNode], Iterable[Union[str, AbstractNode]]]) -> _TD:
        links = {_to_id(parent): set(map(_to_id, children)) for parent, children in links.items()}
        return self._replace(structure=self._structure.without_children(links))

    @property
    def node_links(self) -> Dict[Type[AbstractNodeLink], Set[AbstractNodeLink]]:
        return {
            t: {restore_object(self._id2view[i], self._id2view) for i in ids}
            for t, ids in self._containers[AbstractNodeLink].type2ids.items()
        }

    @_filter_elements(AbstractNodeLink)
    def with_node_links(self: _TD, links: Iterable[AbstractNodeLink], *, update: bool = False) -> _TD:
        return self.with_elements(links, update=update)

    @_filter_elements(AbstractNodeLink)
    def without_node_links(self: _TD, links: Iterable[Union[str, AbstractNodeLink]], *, cascade: bool = False) -> _TD:

        return self.without_elements(map(_to_id, links), cascade=cascade)

    @property
    def facts(self) -> Dict[Type[AbstractFact], Iterable[AbstractFact]]:
        return {
            t: tuple(restore_object(self._id2view[i], self._id2view) for i in ids)
            for t, ids in self._containers[AbstractFact].type2ids.items()
        }

    @_filter_elements(AbstractFact)
    def with_facts(self: _TD, facts: Iterable[AbstractFact], *, update: bool = False) -> _TD:
        return self.with_elements(facts, update=update)

    @_filter_elements(AbstractFact)
    def without_facts(self: _TD, facts: Iterable[Union[str, AbstractFact]], *, cascade: bool = False) -> _TD:
        return self.without_elements(map(_to_id, facts), cascade=cascade)

    @property
    def directives(self) -> Dict[Type[AbstractDirective], Iterable[AbstractDirective]]:
        return {
            t: tuple(restore_object(self._id2view[i], self._id2view) for i in ids)
            for t, ids in self._containers[AbstractDirective].type2ids
        }

    @_filter_elements(AbstractDirective)
    def with_directives(self: _TD, directives: Iterable[AbstractDirective], *, update: bool = False) -> _TD:
        return self.with_elements(directives, update=update)

    @_filter_elements(AbstractDirective)
    def without_directives(self: _TD, directives: Iterable[Union[str, AbstractDirective]], *, cascade: bool = False) -> _TD:
        return self.without_elements(map(_to_id, directives), cascade=cascade)

    @property
    def metadata(self) -> Dict[str, Any]:
        return deepcopy(self._metadata)

    def with_metadata(self: _TD, metadata: Dict[str, Any]) -> _TD:
        return self._replace(metadata=deepcopy(metadata))

    def __hash__(self):
        return hash((self._id, self._containers[AbstractNode], self._containers[AbstractFact]))

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, TalismanDocumentImpl):
            return NotImplemented
        return self._id == o._id and self._id2view == o._id2view and self._containers == o._containers and \
            self._structure == o._structure and self._metadata == o._metadata


def _to_id(obj: Union[str, EnsureIdentifiable]) -> str:
    return obj if isinstance(obj, str) else obj.id


def _filter(
        container: TypedIdsContainer[_D], known_ids: Dict[str, Any], elements: Iterable[Union[str, _D]], base_type: Type[_D]
) -> Iterator[Union[str, _D]]:
    for element in elements:
        if isinstance(element, str):
            if element in known_ids and element not in container:
                raise ValueError(f'Unexpected type for element {element}. Expected: {base_type}')
        else:
            if not isinstance(element, base_type):
                raise ValueError(f'Unexpected type for element {element}. Expected: {base_type}')
            if element.id in known_ids and element.id not in container:
                raise ValueError(f'Identifier collision for {element}')
        yield element



from dataclasses import dataclass

from tdm.future.abstract.datamodel import AbstractNodeMention
from tdm.future.abstract.json_schema import generate_model
from tdm.future.datamodel.nodes import ImageNode


@generate_model
@dataclass(frozen=True)
class ImageNodeMention(AbstractNodeMention):
    node: ImageNode
    top: int
    bottom: int
    left: int
    right: int

    def __post_init__(self):
        if not isinstance(self.node, ImageNode):
            raise ValueError(f"Incorrect node type {type(self.node)}. Expected {ImageNode}")
        if self.top < 0 or self.bottom <= self.top or self.left < 0 or self.right <= self.left:
            raise ValueError(f"Incorrect bbox [({self.top}, {self.left}); ({self.bottom}, {self.right})]")



from typing import Any, Dict, Iterable, Optional, Tuple, Union

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, FactMetadata, FactStatus, FactType

ValueType = Dict[str, Any]  # TODO: maybe we should restrict Any typing

PLATFORM_VALUE = "platform_value"


class ValueFact(AbstractFact[ValueType]):
    def __init__(self, id_: Optional[str], status: FactStatus, type_id: str,
                 value: Optional[Union[ValueType, Tuple[ValueType, ...]]] = None,
                 mention: Iterable[AbstractTalismanSpan] = None,
                 metadata: Optional[FactMetadata] = None):
        super().__init__(id_, FactType.VALUE, status, type_id, value, mention, metadata)

    @AbstractFact.update_metadata
    def with_changes(self: 'ValueFact', *, status: FactStatus = None, type_id: str = None,
                     value: Union[ValueType, Tuple[ValueType, ...]] = None,
                     mention: Tuple[AbstractTalismanSpan, ...] = None,
                     metadata: FactMetadata = None,
                     # metadata fields
                     fact_confidence: float = None,
                     value_confidence: Union[float, Tuple[float, ...]] = None) -> 'ValueFact':
        return ValueFact(
            self._id,
            status if status is not None else self._status,
            type_id if type_id is not None else self._type_id,
            value if value is not None else self._value,
            mention if mention is not None else self._mention,
            metadata if metadata is not None else self._metadata
        )



from tdm.abstract.datamodel import AbstractDirective
from tdm.abstract.datamodel.directive import DirectiveType


class CreatePlatformDirective(AbstractDirective):
    __slots__ = ('_key', '_name', '_platform_type', '_url')

    def __init__(self, key: str, name: str, platform_type: str, url: str):
        super().__init__(DirectiveType.CREATE_PLATFORM)
        self._key = key
        self._name = name
        self._url = url
        self._platform_type = platform_type

    @property
    def key(self) -> str:
        return self._key

    @property
    def platform_type(self) -> str:
        return self._platform_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def url(self) -> str:
        return self._url

    def __eq__(self, other):
        if not isinstance(other, AbstractDirective):
            return NotImplemented
        return isinstance(other, CreatePlatformDirective) and other._key == self._key and other._name == self._name and \
            other._platform_type == self._platform_type and other._url == self._url

    def __hash__(self):
        return hash((super().__hash__(), self._key))



from dataclasses import dataclass
from typing import Sequence, Set, Tuple, Union

from frozendict import frozendict

from tdm.future.abstract.datamodel import AbstractFact, Identifiable
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class _AtomValueFact(AbstractFact):
    type_id: str
    value: Union[dict, Tuple[dict, ...]] = tuple()

    def __post_init__(self):
        if isinstance(self.value, dict):
            object.__setattr__(self, 'value', frozendict(self.value))
        elif isinstance(self.value, Sequence):
            object.__setattr__(self, 'value', tuple(frozendict(v) for v in self.value))

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}


@generate_model(label='atom')
@dataclass(frozen=True)
class AtomValueFact(Identifiable, _AtomValueFact):
    pass


@dataclass(frozen=True)
class _CompositeValueFact(AbstractFact):
    type_id: str

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}


@generate_model(label='composite')
@dataclass(frozen=True)
class CompositeValueFact(Identifiable, _CompositeValueFact):
    pass



from tdm.future.abstract.datamodel import AbstractFact, AbstractNode, AbstractNodeMention, BaseNodeMetadata
from .identifiable import IdSerializer
from .mention import NodeMentionSerializer
from .metadata import NodeMetadataSerializer


def build_serializers():
    result = {
        AbstractNode: IdSerializer(AbstractNode),
        AbstractFact: IdSerializer(AbstractFact),
        AbstractNodeMention: NodeMentionSerializer(),
        BaseNodeMetadata: NodeMetadataSerializer()
    }
    # other serializers could be added here
    return result



from dataclasses import dataclass
from typing import Optional

from tdm.future.abstract.datamodel import AbstractContentNode, BaseNodeMetadata
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class Base64NodeMetadata(BaseNodeMetadata):
    content_type: Optional[str] = None


@generate_model(label='base64')
@dataclass(frozen=True)
class Base64Node(AbstractContentNode[Base64NodeMetadata, str]):
    pass



from dataclasses import dataclass
from typing import Set

from tdm.future.abstract.datamodel import AbstractFact, AbstractNodeMention, Identifiable
from tdm.future.abstract.json_schema import generate_model
from .value import AtomValueFact


@dataclass(frozen=True)
class _MentionFact(AbstractFact):
    mention: AbstractNodeMention
    value: AtomValueFact

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'mention', 'value'}


@generate_model(label='mention')
@dataclass(frozen=True)
class MentionFact(Identifiable, _MentionFact):
    pass



from dataclasses import dataclass

from tdm.future.abstract.datamodel import AbstractNodeMention
from tdm.future.abstract.json_schema import generate_model


@generate_model
@dataclass(frozen=True)
class NodeMention(AbstractNodeMention):
    pass



from abc import ABCMeta
from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from typing import Generic, Optional, Set, TypeVar

from .base import EnsureIdentifiable


@total_ordering
class FactStatus(str, Enum):
    def __new__(cls, name: str, priority: int):
        obj = str.__new__(cls, name)
        obj._value_ = name
        obj.priority = priority
        return obj

    APPROVED = ("approved", 0)
    DECLINED = ("declined", 1)
    AUTO = ("auto", 2)
    HIDDEN = ("hidden", 3)
    NEW = ("new", 4)

    def __lt__(self, other: 'FactStatus'):
        if not isinstance(other, FactStatus):
            return NotImplemented
        return self.priority < other.priority


@dataclass(frozen=True)
class AbstractFact(EnsureIdentifiable, metaclass=ABCMeta):
    status: FactStatus


_ST = TypeVar('_ST', bound=AbstractFact)
_TT = TypeVar('_TT', bound=AbstractFact)


@dataclass(frozen=True)
class AbstractLinkFact(AbstractFact, Generic[_ST, _TT]):
    type_id: str
    source: _ST
    target: _TT
    value: Optional[str] = None

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'source', 'target', 'type_id'}



from dataclasses import dataclass
from typing import Generic, Set, TypeVar

from .base import EnsureIdentifiable
from .mention import AbstractNodeMention

_ST = TypeVar('_ST', bound=AbstractNodeMention)
_TT = TypeVar('_TT', bound=AbstractNodeMention)


@dataclass(frozen=True)
class AbstractNodeLink(EnsureIdentifiable, Generic[_ST, _TT]):
    source: _ST
    target: _TT

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'source', 'target'}



__all__ = [
    'AbstractElementModel', 'AbstractElementSerializer', 'AbstractModelSerializer',
    'IdSerializer',
    'NodeMentionSerializer',
    'NodeMetadataSerializer',
    'BaseSerializers'
]

from typing import Dict

from frozendict import frozendict

from .abstract import AbstractElementModel, AbstractElementSerializer, AbstractModelSerializer
from .identifiable import IdSerializer
from .mention import NodeMentionSerializer
from .metadata import NodeMetadataSerializer
from .serializers import build_serializers

BaseSerializers: Dict[type, AbstractElementSerializer] = frozendict(build_serializers())



from typing import Any, Dict, Type, TypeVar

from tdm.future.abstract.datamodel import EnsureIdentifiable
from .abstract import AbstractElementSerializer

_Identifiable = TypeVar('_Identifiable', bound=EnsureIdentifiable)


class IdSerializer(AbstractElementSerializer[_Identifiable, str]):
    def __init__(self, type_: Type[_Identifiable]):
        self._type = type_

    def serialize(self, element: _Identifiable) -> str:
        return element.id

    def deserialize(self, serialized: str, typed_id2element: Dict[type, Dict[str, Any]]) -> _Identifiable:
        return typed_id2element[self._type][serialized]

    def field_type(self, element_type: Type[_Identifiable]) -> Type[str]:
        return str



from dataclasses import dataclass
from typing import Optional

from tdm.future.abstract.datamodel import AbstractContentNode, BaseNodeMetadata
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class FileNodeMetadata(BaseNodeMetadata):
    name: Optional[str] = None
    size: Optional[int] = None


@generate_model(label='file')
@dataclass(frozen=True)
class FileNode(AbstractContentNode[FileNodeMetadata, str]):
    pass


@dataclass(frozen=True)
class ImageNodeMetadata(FileNodeMetadata):
    width: Optional[int] = None
    height: Optional[int] = None


@generate_model(label='image')
@dataclass(frozen=True)
class ImageNode(AbstractContentNode[ImageNodeMetadata, str]):
    pass



__all__ = [
    'CreateAccountDirective',
    'CreateConceptDirective',
    'CreatePlatformDirective'
]

from .account import CreateAccountDirective
from .concept import CreateConceptDirective
from .platform import CreatePlatformDirective



from dataclasses import dataclass
from typing import Optional, Tuple

from frozendict import frozendict

from tdm.future.abstract.datamodel import AbstractDirective, Identifiable
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class _CreateConceptDirective(AbstractDirective):
    name: str
    concept_type: str
    filters: Tuple[dict, ...]
    notes: Optional[str] = None
    markers: Optional[str] = None
    access_level: Optional[str] = None

    def __post_init__(self):
        filters = tuple(frozendict(f) for f in self.filters)
        object.__setattr__(self, 'filters', filters)


@generate_model(label='create_concept')
@dataclass(frozen=True)
class CreateConceptDirective(Identifiable, _CreateConceptDirective):
    pass



from typing import Dict, Type, Union

from tdm.future.abstract.datamodel import AbstractDirective, AbstractFact, AbstractNode, AbstractNodeLink, EnsureIdentifiable

_BASE_TYPES = {AbstractNode, AbstractNodeLink, AbstractFact, AbstractDirective}

_CACHE: Dict[Type[EnsureIdentifiable], Type[EnsureIdentifiable]] = {}


def get_base_type(element: Union[EnsureIdentifiable, Type[EnsureIdentifiable]]) -> Type[EnsureIdentifiable]:
    if isinstance(element, EnsureIdentifiable):
        element = type(element)
    if element in _CACHE:
        return _CACHE[element]
    bases = {t for t in _BASE_TYPES if issubclass(element, t)}
    if len(bases) != 1:
        raise TypeError(f"Unsupported element type {element}")
    _CACHE[element] = bases.pop()
    return _CACHE[element]



from typing import Any, Callable, Dict, Generic, Optional, Tuple, TypeVar, Union


def unfold_union(type_: type) -> Tuple[type, ...]:
    # check if it is union
    if hasattr(type_, '__origin__') and type_.__origin__ is Union:
        return type_.__args__
    return type_,


def is_subclass(type_: type, base_type: type) -> bool:
    try:
        return all(issubclass(t, base_type) for t in unfold_union(type_))
    except TypeError:
        return False  # here we should think about List[base_type] and others


def cache_result(key_param: Optional[str] = None, cache: Optional[dict] = None):
    if cache is None:
        cache = {}

    def wrap(f: Callable):
        def wrapper(*args, **kwargs) -> Any:
            if key_param is not None:
                key = kwargs[key_param]
            elif args:
                key = args[0]
            else:
                key = kwargs[next(iter(kwargs))]
            if key in cache:
                return cache[key]
            result = f(*args, **kwargs)
            cache[key] = result
            return result

        return wrapper

    return wrap


def generics_mapping(type_: type) -> Dict[TypeVar, type]:
    if issubclass(type_, Generic):  # actually it works (should be rewritten in more pythonic way)
        # assume type vars are not intersect
        result = {}
        for orig_base in type_.__orig_bases__:
            if '__origin__' in orig_base.__dict__ and '__args__' in orig_base.__dict__:
                result.update(dict(zip(orig_base.__origin__.__parameters__, orig_base.__args__)))
        return result
    return {}



import dataclasses
from dataclasses import fields
from typing import Any, Dict, Generic, Optional, Type, TypeVar

from pydantic import BaseModel, Extra, create_model, root_validator
from typing_extensions import Literal, Self

from tdm.future.helper import cache_result, generics_mapping, is_subclass
from .serializers import AbstractElementModel, AbstractElementSerializer, BaseSerializers

_Element = TypeVar('_Element')


# hackish root validator solution as normal validators are not called for skipped required fields
def set_type_if_none(cls, values):
    if 'type' not in values:
        values['type'] = cls.__fields__['type'].type_.__args__[0]  # get arg of Literal ('type' field type)
    return values


class ElementModel(BaseModel, AbstractElementModel[_Element], Generic[_Element]):
    class Config:
        extra = Extra.forbid

    def deserialize(self, typed_id2element: Dict[type, Dict[str, Any]]) -> _Element:
        raise NotImplementedError

    @classmethod
    def serialize(cls, element: _Element) -> Self:
        kwargs = {}
        for key, value in element.__dict__.items():
            for type_, serializer in BaseSerializers.items():
                if isinstance(value, type_):
                    kwargs[key] = serializer.serialize(value)
                    break
            else:
                kwargs[key] = value
        return cls(**kwargs)


_BaseType = TypeVar('_BaseType', bound=type)


@cache_result()
def create_model_for_type(type_: Type[_Element], label: Optional[str] = None) -> Type[ElementModel[_Element]]:
    type_vars = generics_mapping(type_)

    model_fields = {}
    validators = {}

    special_fields: Dict[str, AbstractElementSerializer] = {}

    for field in fields(type_):
        name = field.name
        default_value = field.default if field.default is not dataclasses.MISSING else ...
        field_type = type_vars.get(field.type, field.type)
        for _type, serializer in BaseSerializers.items():
            if is_subclass(field_type, _type):
                field_type = serializer.field_type(field_type)
                special_fields[name] = serializer
        model_fields[name] = (field_type, default_value)
    if label:
        model_fields['type'] = (Literal[label], ...)
        validators['set_if_none'] = root_validator(pre=True, allow_reuse=True)(set_type_if_none)

    model = create_model(f"{type_.__name__}Model", __base__=ElementModel, __validators__=validators, **model_fields)

    def deserialize(self: model, typed_id2element: Dict[type, Dict[str, Any]]) -> type_:
        kwargs = {}
        for f in set(type_.__dataclass_fields__).intersection(self.__dict__):
            kwargs[f] = getattr(self, f)
            if f in special_fields:
                kwargs[f] = special_fields[f].deserialize(kwargs[f], typed_id2element)
        return type_(**kwargs)

    model.deserialize = deserialize

    return model



from dataclasses import dataclass
from typing import Sequence, Set, Tuple, Union

from tdm.future.abstract.datamodel import AbstractFact, Identifiable
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class _ConceptFact(AbstractFact):  # not an error as id argument is kw only
    type_id: str
    value: Union[str, Tuple[str, ...]] = tuple()

    def __post_init__(self):
        if isinstance(self.value, str) or isinstance(self.value, tuple):
            return
        if isinstance(self.value, Sequence):
            object.__setattr__(self, 'value', tuple(self.value))
        else:
            raise ValueError

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}


@generate_model(label='concept')
@dataclass(frozen=True)
class ConceptFact(Identifiable, _ConceptFact):
    pass



from dataclasses import dataclass
from typing import Union

from tdm.future.abstract.datamodel import AbstractLinkFact, Identifiable
from tdm.future.abstract.json_schema import generate_model
from .concept import ConceptFact
from .value import AtomValueFact, CompositeValueFact


@generate_model(label='relation')
@dataclass(frozen=True)
class RelationFact(Identifiable, AbstractLinkFact[ConceptFact, ConceptFact]):
    pass


ValueFact = Union[AtomValueFact, CompositeValueFact]


@generate_model(label='property')
@dataclass(frozen=True)
class PropertyFact(Identifiable, AbstractLinkFact[ConceptFact, ValueFact]):
    pass


@generate_model(label='r_property')
@dataclass(frozen=True)
class RelationPropertyFact(Identifiable, AbstractLinkFact[RelationFact, ValueFact]):
    pass


@generate_model(label='slot')
@dataclass(frozen=True)
class SlotFact(Identifiable, AbstractLinkFact[CompositeValueFact, ValueFact]):
    pass



import uuid
from typing import Dict, Iterable, Optional, Tuple

from tdm.abstract.datamodel.document import DocumentMetadata
from tdm.future.abstract.datamodel import AbstractDirective, AbstractFact, AbstractNode, AbstractNodeLink
from tdm.future.abstract.datamodel.document import AbstractDocumentFactory, TalismanDocument
from ._container import TypedIdsContainer
from ._impl import TalismanDocumentImpl
from ._structure import NodesStructure


class TalismanDocumentFactory(AbstractDocumentFactory):

    def create_document(self, *, id_: Optional[str] = None) -> TalismanDocument:
        return TalismanDocumentImpl(
            id2view={},
            dependencies={},
            structure=NodesStructure(),
            containers={
                AbstractNode: TypedIdsContainer(AbstractNode, ()),
                AbstractNodeLink: TypedIdsContainer(AbstractNodeLink, ()),
                AbstractFact: TypedIdsContainer(AbstractFact, ()),
                AbstractDirective: TypedIdsContainer(AbstractDirective, ())
            },
            metadata=None,
            id_=id_ or self.generate_id())

    def construct(
            self,
            content: Iterable[AbstractNode] = (),
            structure: Dict[str, Tuple[str, ...]] = None,
            root: Optional[str] = None,
            node_links: Iterable[AbstractNodeLink] = (),
            facts: Iterable[AbstractFact] = (),
            directives: Iterable[AbstractDirective] = (),
            metadata: Optional[DocumentMetadata] = None,
            *, id_: Optional[str] = None
    ) -> TalismanDocument:
        doc: TalismanDocumentImpl = self.create_document(id_=id_)
        doc = doc.with_elements((*content, *node_links, *facts, *directives)) \
            .with_links(structure) \
            .with_main_root(root) \
            .with_metadata(metadata)
        return doc

    @staticmethod
    def generate_id():
        return str(uuid.uuid4())



from typing import Any, Dict

from typing_extensions import Literal

from tdm.abstract.datamodel import AbstractFact, FactType
from tdm.datamodel.fact import ValueFact
from tdm.json_schema.fact.common import AbstractFactModel, FactFactory


class ValueFactModel(AbstractFactModel):
    fact_type: Literal[FactType.VALUE] = FactType.VALUE

    def to_value(self, mapping: Dict[str, AbstractFact]) -> Any:
        return self.value

    @property
    def fact_factory(self) -> FactFactory:
        return ValueFact

    @classmethod
    def build_value(cls, value: Any) -> Any:
        return value



from dataclasses import dataclass
from typing import Optional

from tdm.future.abstract.datamodel import AbstractContentNode, BaseNodeMetadata
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class TextNodeMetadata(BaseNodeMetadata):
    language: Optional[str] = None


@generate_model(label='text')
@dataclass(frozen=True)
class TextNode(AbstractContentNode[TextNodeMetadata, str]):
    pass


@generate_model(label='key')
@dataclass(frozen=True)
class KeyNode(AbstractContentNode[TextNodeMetadata, str]):
    pass



__all__ = [
    'FactStatus',
    'ConceptFact',
    'PropertyFact', 'RelationFact', 'RelationPropertyFact', 'SlotFact',
    'MentionFact',
    'AtomValueFact', 'CompositeValueFact',
]

from tdm.abstract.datamodel import FactStatus
from .concept import ConceptFact
from .links import PropertyFact, RelationFact, RelationPropertyFact, SlotFact
from .mention import MentionFact
from .value import AtomValueFact, CompositeValueFact



from typing import Type

from tdm.future.abstract.datamodel import AbstractNodeMention
from .abstract import AbstractElementModel, AbstractModelSerializer


class NodeMentionSerializer(AbstractModelSerializer[AbstractNodeMention]):
    def serialize(self, element: AbstractNodeMention) -> AbstractElementModel[AbstractNodeMention]:
        from tdm.future.json_schema.mentions import serialize_mention
        return serialize_mention(element)

    def field_type(self, element_type: Type[AbstractNodeMention]) -> Type[AbstractElementModel[AbstractNodeMention]]:
        from tdm.future.json_schema.mentions import MENTION_MODELS
        return MENTION_MODELS[element_type]



from dataclasses import dataclass

from tdm.future.abstract.datamodel import AbstractNodeMention
from tdm.future.abstract.json_schema import generate_model
from tdm.future.datamodel.nodes import TextNode


@generate_model
@dataclass(frozen=True)
class TextNodeMention(AbstractNodeMention):
    node: TextNode
    start: int
    end: int

    def __post_init__(self):
        if not isinstance(self.node, TextNode):
            raise ValueError(f"Incorrect node type {type(self.node)}. Expected {TextNode}")
        if self.start < 0 or self.end <= self.start:
            raise ValueError(f"Incorrect span [{self.start}, {self.end})")



__all__ = [
    'AbstractLabeledModel', 'ModelsGenerator',
    'generate_model', 'get_model_generator',
    'ElementModel'
]

from .composite import AbstractLabeledModel, ModelsGenerator
from .decorator import generate_model, get_model_generator
from .model import ElementModel



from dataclasses import dataclass, field
from typing import Generic, Set, TypeVar

from frozendict import frozendict

from tdm.future.helper import generics_mapping
from .base import Identifiable


@dataclass(frozen=True)
class BaseNodeMetadata:
    hidden: bool = False


_NodeMetadata = TypeVar("_NodeMetadata", bound=BaseNodeMetadata)


@dataclass(frozen=True)
class AbstractNode(Identifiable, Generic[_NodeMetadata]):
    metadata: _NodeMetadata = None

    def __post_init__(self):
        super().__post_init__()
        if self.metadata is None:
            # hack for runtime metadata generation (if no value passed)
            object.__setattr__(self, 'metadata', self._generate_metadata())

    def _generate_metadata(self) -> _NodeMetadata:
        type_vars = generics_mapping(type(self))
        return type_vars[self.__dataclass_fields__['metadata'].type]()

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return set()


_Content = TypeVar('_Content')


@dataclass(frozen=True)
class _AbstractContentNode(Generic[_Content]):
    content: _Content
    markup: dict = field(default_factory=frozendict)

    def __post_init__(self):
        object.__setattr__(self, 'markup', frozendict(self.markup))


@dataclass(frozen=True)
class AbstractContentNode(AbstractNode[_NodeMetadata], _AbstractContentNode[_Content], Generic[_NodeMetadata, _Content]):

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'content'}



from typing import Type, TypeVar

from tdm.future.abstract.datamodel import AbstractDirective, AbstractFact, AbstractNode, AbstractNodeLink, AbstractNodeMention
from .composite import ModelsGenerator

_MODELS_GENERATORS = {
    AbstractDirective: ModelsGenerator(AbstractDirective),
    AbstractFact: ModelsGenerator(AbstractFact),
    AbstractNodeLink: ModelsGenerator(AbstractNodeLink),
    AbstractNodeMention: ModelsGenerator(AbstractNodeMention, include_label=True),
    AbstractNode: ModelsGenerator(AbstractNode, include_label=True)
}

_Element = TypeVar('_Element')


def get_model_generator(cls: Type[_Element]) -> ModelsGenerator[_Element]:
    for cls_, generator in _MODELS_GENERATORS.items():
        if issubclass(cls, cls_):
            return generator
    raise TypeError(f"There is no model generator for {cls}. Available: {tuple(_MODELS_GENERATORS)}")


def generate_model(cls: type = None, *, label: str = None):
    def wrapper(cls_: type):
        return get_model_generator(cls_).generate_model(cls_, label=label)

    if cls is None:
        return wrapper
    return wrapper(cls)



from typing import Type

from tdm.future.abstract.datamodel import AbstractNodeLink
from tdm.future.abstract.json_schema import AbstractLabeledModel, get_model_generator


def register_node_link_models() -> Type[AbstractLabeledModel[AbstractNodeLink]]:
    import tdm.future.datamodel.links as links
    links

    return get_model_generator(AbstractNodeLink).generate_labeled_model('NodeLinksModel')


NodeLinksModel = register_node_link_models()



from dataclasses import dataclass
from typing import Optional

from tdm.future.abstract.datamodel import AbstractNode, BaseNodeMetadata
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class ListNodeMetadata(BaseNodeMetadata):
    bullet: Optional[str] = None


@generate_model(label='list')
@dataclass(frozen=True)
class ListNode(AbstractNode[ListNodeMetadata]):
    pass


@generate_model(label='json')
@dataclass(frozen=True)
class JSONNode(AbstractNode[BaseNodeMetadata]):
    pass


@generate_model(label='table')
@dataclass(frozen=True)
class TableNode(AbstractNode[BaseNodeMetadata]):
    pass


@dataclass(frozen=True)
class TableRowNodeMetadata(BaseNodeMetadata):
    header: Optional[bool] = None


@generate_model(label='row')
@dataclass(frozen=True)
class TableRowNode(AbstractNode[TableRowNodeMetadata]):
    pass


@dataclass(frozen=True)
class TableCellNodeMetadata(BaseNodeMetadata):
    colspan: Optional[int] = None
    rowspan: Optional[int] = None


@generate_model(label='cell')
@dataclass(frozen=True)
class TableCellNode(AbstractNode[TableCellNodeMetadata]):
    pass



from dataclasses import dataclass, field

from tdm.future.abstract.datamodel import AbstractDirective, Identifiable
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class _CreatePlatformDirective(AbstractDirective):
    key: str
    name: str
    platform_type: str
    url: str


@generate_model(label='create_platform')
@dataclass(frozen=True)
class CreatePlatformDirective(Identifiable, _CreatePlatformDirective):
    id: str = field(default=None, compare=False)



from typing import Dict, Optional

from pydantic import BaseModel

from tdm.abstract.datamodel import NodeType
from tdm.datamodel import TreeDocumentContent


class NodeMetadata(BaseModel):
    node_type: NodeType
    original_text: Optional[str]
    text_translations: Dict[str, str] = {}
    language: Optional[str] = None
    hidden: bool = False

    class Config:
        extra = 'allow'  # any other extra fields will be kept

    def parse(self) -> dict:
        return self.dict()

    @classmethod
    def build(cls, node: TreeDocumentContent) -> 'NodeMetadata':
        return cls.construct(
            node_type=node.type,
            original_text=node.original_node_text if node.original_node_text != node.node_text else None,
            text_translations={lang: node.get_node_text_translation(lang) for lang in node.node_text_languages},
            hidden=node.is_hidden,
            language=node.language
        )



import uuid
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Set


@dataclass(frozen=True)
class EnsureIdentifiable(object):  # we need such ugly inheritance to guarantee default valued fields follows fields without defaults

    def __post_init__(self):
        if not isinstance(self, Identifiable):
            raise TypeError(f"{type(self)} should inherit {Identifiable}. Actual mro is {type(self).mro()}")

    @classmethod
    @abstractmethod
    def constant_fields(cls) -> Set[str]:
        pass


@dataclass(frozen=True)
class Identifiable(EnsureIdentifiable, metaclass=ABCMeta):
    id: str = field(default_factory=lambda: None)

    def __post_init__(self):
        if self.id is None:
            object.__setattr__(self, 'id', self.generate_id())
        for type_ in type(self).mro():
            if issubclass(type_, Identifiable):
                continue
            if hasattr(type_, '__post_init__'):
                type_.__post_init__(self)

    @staticmethod
    def generate_id() -> str:
        return str(uuid.uuid4())



__all__ = [
    'Base64Node', 'Base64NodeMetadata',
    'FileNode', 'FileNodeMetadata', 'ImageNode', 'ImageNodeMetadata',
    'JSONNode', 'ListNode', 'ListNodeMetadata', 'TableCellNode', 'TableCellNodeMetadata', 'TableNode', 'TableRowNode',
    'TableRowNodeMetadata',
    'KeyNode', 'TextNode', 'TextNodeMetadata'
]

from .base64 import Base64Node, Base64NodeMetadata
from .file import FileNode, FileNodeMetadata, ImageNode, ImageNodeMetadata
from .structure import JSONNode, ListNode, ListNodeMetadata, TableCellNode, TableCellNodeMetadata, TableNode, TableRowNode, \
    TableRowNodeMetadata
from .text import KeyNode, TextNode, TextNodeMetadata



from collections import defaultdict
from itertools import chain
from typing import Dict, Generic, Iterable, Iterator, Set, Tuple, Type, TypeVar, Union

from frozendict import frozendict

_T = TypeVar('_T')


class TypedIdsContainer(Iterable[Tuple[str, type]], Generic[_T]):
    __slots__ = ('_base', '_id2type', '_type2ids')

    def __init__(self, base: Type[_T], data: Union[Dict[str, Type[_T]], Iterable[Tuple[str, Type[_T]]]]):
        self._base = base
        if not isinstance(data, dict):
            data = dict(data)
        if any(not issubclass(t, self._base) for t in data.values()):
            raise ValueError
        self._id2type = frozendict(data)
        # self._id2data = frozendict(data if isinstance(data, dict) else {item.id: item for item in data})
        self._type2ids = None

    @classmethod
    def create(cls, base: Type[_T], data: Iterable[Tuple[str, type]]):
        return data if isinstance(data, cls) and data._base is base else cls(base, data)

    def __iter__(self) -> Iterator[Tuple[str, type]]:
        return iter(self._id2type.items())

    @property
    def ids(self) -> Set[str]:
        return set(self._id2type)

    @property
    def type2ids(self) -> Dict[type, Iterable[str]]:
        if self._type2ids is None:
            result = defaultdict(list)
            for i, t in self._id2type.items():
                result[t].append(i)
            self._type2ids = {t: frozenset(ids) for t, ids in result.items()}
        return self._type2ids

    def with_ids(self, data: Iterable[Tuple[str, type]]):
        return TypedIdsContainer(self._base, chain(self._id2type.items(), data))

    def without_ids(self, data: Iterable[str]):
        ids = set(data)
        return TypedIdsContainer(self._base, filter(lambda e: e[0] not in ids, self._id2type.items()))

    def __contains__(self, item: Union[str, Tuple[str, type]]) -> bool:
        if isinstance(item, str):
            return item in self._id2type
        return self._id2type.get(item[0], None) == item[1]

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, TypedIdsContainer):
            return NotImplemented
        return self._base is o._base and self._id2type == o._id2type

    def __hash__(self) -> int:
        return hash((self._base, self._id2type))



from typing import Callable, Iterable, Tuple, Type, Union

from pydantic import Field, create_model
from typing_extensions import Annotated

from tdm.future.abstract.datamodel import AbstractNode, TalismanDocument
from tdm.future.abstract.json_schema import ElementModel, get_model_generator
from tdm.future.helper import unfold_union


def register_node_models() -> Tuple[Type[ElementModel[AbstractNode]], Callable[[AbstractNode], ElementModel[AbstractNode]]]:
    import tdm.future.datamodel.nodes as nodes  # we need this import for serializers registration
    nodes

    # TODO: here plugin for extra document nodes could be added

    models, serialize = get_model_generator(AbstractNode).generate_union_model()

    # add children field to all node models. serialize function doesn't generate value for children field
    kwargs = {
        'children': (Tuple[str, ...], ...)
    }

    model_mapping = {}
    tree_models = []

    for model in unfold_union(models[AbstractNode]):
        tree_model = create_model(f"Tree{model.__name__}", __base__=model, **kwargs)
        model_mapping[model] = tree_model
        tree_models.append(tree_model)

    NodeModel_ = Annotated[Union[tuple(tree_models)], Field(discriminator='type')]  # noqa N806

    # wrap serialize to generate tree node instead of node
    def serialize_node(node: AbstractNode) -> ElementModel[AbstractNode]:
        serialized = serialize(node)
        return model_mapping[type(serialized)].construct(**serialized.__dict__)

    return NodeModel_, serialize_node


NodeModel, serialize_node = register_node_models()


def fill_children(node_models: Iterable[NodeModel], document: TalismanDocument) -> None:
    for node in node_models:
        children = tuple(n.id for n in document.child_nodes(node.id))
        node.children = children



from typing import Type

from tdm.future.abstract.datamodel import BaseNodeMetadata
from .abstract import AbstractElementModel, AbstractModelSerializer


class NodeMetadataSerializer(AbstractModelSerializer[BaseNodeMetadata]):
    def serialize(self, element: BaseNodeMetadata) -> AbstractElementModel[BaseNodeMetadata]:
        return self.field_type(type(element)).serialize(element)

    def field_type(self, element_type: Type[BaseNodeMetadata]) -> Type[AbstractElementModel[BaseNodeMetadata]]:
        from tdm.future.abstract.json_schema.model import create_model_for_type
        return create_model_for_type(element_type)  # actually it is cached



from dataclasses import dataclass

from tdm.future.abstract.datamodel import AbstractNodeLink, AbstractNodeMention, Identifiable
from tdm.future.abstract.json_schema import generate_model


@generate_model(label='same')
@dataclass(frozen=True)
class SameNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    pass


@generate_model(label='translation')
@dataclass(frozen=True)
class TranslationNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    pass


@generate_model(label='reference')
@dataclass(frozen=True)
class ReferenceNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    pass



from dataclasses import dataclass

from tdm.future.abstract.datamodel import AbstractDirective, Identifiable
from tdm.future.abstract.json_schema import generate_model


@dataclass(frozen=True)
class _CreateAccountDirective(AbstractDirective):
    key: str
    platform_key: str
    name: str
    url: str


@generate_model(label='create_account')
@dataclass(frozen=True)
class CreateAccountDirective(Identifiable, _CreateAccountDirective):
    pass



from abc import abstractmethod
from collections import defaultdict
from typing import Callable, Dict, Generic, Iterable, List, Tuple, Type, TypeVar, Union

from pydantic import BaseModel, Field, create_model
from typing_extensions import Annotated, Self

from tdm.future.abstract.datamodel import AbstractNode
from .model import ElementModel, create_model_for_type

_ElementType = TypeVar('_ElementType')
_DATA = TypeVar('_DATA')


class AbstractLabeledModel(Generic[_ElementType]):

    @abstractmethod
    def deserialize(self, nodes: Dict[str, AbstractNode]) -> Iterable[_ElementType]:
        pass

    @classmethod
    @abstractmethod
    def serialize(cls, elements: Dict[Type[_ElementType], Iterable[_ElementType]]) -> Self:
        pass


class ModelsGenerator(Generic[_DATA]):
    __slots__ = (
        '_base_type', '_include_label',
        '_finalized', '_label2model', '_cls2label_model', '_cls_hierarchy'
    )

    def __init__(self, base_type: Type[_DATA], include_label: bool = False):
        self._base_type = base_type
        self._include_label = include_label

        self._finalized = False
        self._label2model: Dict[str, Type[ElementModel[_DATA]]] = {}
        self._cls2label_model: Dict[Type[_DATA], Tuple[str, Type[ElementModel[_DATA]]]] = {}
        self._cls_hierarchy: Dict[Type[_DATA], List[Type[_DATA]]] = defaultdict(list)

    def generate_model(self, cls: Type[_DATA] = None, *, label: str = None):
        def wrap(cls: Type[_DATA]):
            label_ = label if label is not None else cls.__name__
            if label_ in self._label2model or cls in self._cls2label_model:
                raise ValueError
            self._register(label_, cls)
            return cls

        if cls is None:
            return wrap

        return wrap(cls)

    def _register(self, label: str, cls: Type[_DATA]) -> None:
        if self._finalized:
            raise Exception
        model = create_model_for_type(cls, label if self._include_label else None)
        self._label2model[label] = model
        self._cls2label_model[cls] = label, model
        for t in cls.mro():
            if issubclass(t, self._base_type):
                self._cls_hierarchy[t].append(cls)

    def generate_labeled_model(self, name: str) -> Type[AbstractLabeledModel[_DATA]]:
        if self._finalized:
            raise Exception
        self._finalized = True
        deserialization_order = tuple(self._label2model)

        base_type = self._base_type

        class LabeledModel(BaseModel, AbstractLabeledModel[_DATA]):

            def deserialize(self, nodes: Dict[str, AbstractNode]) -> Iterable[_ElementType]:
                id2element: Dict[str, _DATA] = {}
                elements = {
                    AbstractNode: nodes,
                    base_type: id2element
                }
                for label in deserialization_order:
                    for model in self.__dict__.get(label) or tuple():
                        model: ElementModel
                        element = model.deserialize(elements)
                        id2element[element.id] = element
                return id2element.values()

            @classmethod
            def serialize(cls, elements: Dict[Type[_ElementType], Iterable[_ElementType]]) -> Self:
                kwargs = {}
                for type_, els in elements.items():
                    label, model = self._cls2label_model[type_]
                    kwargs[label] = tuple(model.serialize(element) for element in els)
                return cls.construct(**kwargs)

        model_fields = {label: (Tuple[model, ...], None) for label, model in self._label2model.items()}
        model = create_model(name, __base__=LabeledModel, **model_fields)
        return model

    def generate_union_model(self, discriminator: str = None) \
            -> Tuple[Dict[Type[_DATA], Type[ElementModel[_DATA]]], Callable[[_DATA], ElementModel[_DATA]]]:
        if self._finalized:
            raise ValueError
        self._finalized = True

        result = {}
        for cls, clss in self._cls_hierarchy.items():
            if len(clss) == 1:
                result[cls] = self._cls2label_model[clss[0]][1]
            else:
                types = tuple(map(lambda cls_: self._cls2label_model[cls_][1], clss))
                result[cls] = Annotated[Union[types], Field(discriminator=discriminator)] if discriminator is not None else Union[types]

        def serialize(element: _DATA) -> ElementModel[_DATA]:
            return result[type(element)].serialize(element)

        return result, serialize



from typing import Type

from tdm.future.abstract.datamodel import AbstractDirective
from tdm.future.abstract.json_schema import AbstractLabeledModel, get_model_generator


def register_directives_model() -> Type[AbstractLabeledModel[AbstractDirective]]:
    import tdm.future.datamodel.directives as directives
    directives

    # TODO: here plugin for extra document nodes could be added

    return get_model_generator(AbstractDirective).generate_labeled_model('DirectivesModel')


DirectivesModel = register_directives_model()



from abc import ABCMeta
from functools import total_ordering
from typing import Tuple

from tdm.abstract.datamodel import AbstractSpan, AbstractTalismanSpan


class DefaultSpan(AbstractSpan, metaclass=ABCMeta):
    __slots__ = ('_start_idx', '_end_idx')

    def __init__(self, start_idx: int, end_idx: int, *, nullable=False):
        self._start_idx = start_idx
        self._end_idx = end_idx

        if not (0 <= self._start_idx <= self._end_idx):
            raise Exception("Incorrect span range")
        if not nullable and self._start_idx == self._end_idx:
            raise Exception("Incorrect span range")

    @property
    def start_idx(self) -> int:
        return self._start_idx

    @property
    def end_idx(self) -> int:
        return self._end_idx

    @property
    def length(self) -> int:
        return self._end_idx - self._start_idx


class NullableSpan(DefaultSpan):
    __slots__ = ()

    def __init__(self, start: int, end: int):
        super().__init__(start, end, nullable=True)

    def shift(self, shift: int) -> 'NullableSpan':
        return NullableSpan(self._start_idx + shift, self._end_idx + shift)

    def _as_tuple(self) -> Tuple[int, int]:
        return self.start_idx, self.end_idx

    def __eq__(self, other):
        if not isinstance(other, NullableSpan):
            return NotImplemented
        return other._as_tuple() == self._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    def __repr__(self):
        return repr(self._as_tuple())

    def __lt__(self, other):
        if not isinstance(other, NullableSpan):
            return NotImplemented
        return self._as_tuple() < other._as_tuple()


def _compare_ids(func):
    def decorate(self: AbstractTalismanSpan, obj: AbstractSpan):
        if not isinstance(obj, AbstractTalismanSpan) or self.node_id != obj.node_id:
            return False
        return func(self, obj)
    return decorate


@total_ordering
class TalismanSpan(DefaultSpan, AbstractTalismanSpan):
    __slots__ = ('_id', )

    def __init__(self, id_: str, start_idx: int, end_idx: int):
        super().__init__(start_idx, end_idx, nullable=True)
        self._id = id_

    @property
    def node_id(self):
        return self._id

    def _as_tuple(self) -> Tuple[str, int, int]:
        return self._id, self.start_idx, self.end_idx

    def shift(self, shift: int) -> 'TalismanSpan':
        return TalismanSpan(self._id, self._start_idx + shift, self._end_idx + shift)

    coincides = _compare_ids(DefaultSpan.coincides)
    intersects = _compare_ids(DefaultSpan.intersects)
    contains = _compare_ids(DefaultSpan.contains)

    def distance(self, obj: 'AbstractSpan') -> int:
        if not isinstance(obj, AbstractTalismanSpan) or obj.node_id != self.node_id:
            raise ValueError(f'Spans {self} and {obj} are on the different nodes!')
        return super().distance(obj)

    def __eq__(self, other):
        if not isinstance(other, TalismanSpan):
            return NotImplemented
        return other._as_tuple() == self._as_tuple()

    def __lt__(self, other):
        if not isinstance(other, TalismanSpan):
            return NotImplemented
        if self.node_id != other.node_id:
            raise ValueError(f'Spans {self} and {other} are on the different nodes!')
        return self._as_tuple() < other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    def __repr__(self):
        return repr(self._as_tuple())
